% *************************************************************
% author:   Gene Cheung                                      **
% date:     09/28/2012                                       **
% modified: 11/05/2012                                       **
% purpose:  find least-square fitting line on 2D plot        **
% *************************************************************

% given direction code, find line-of-best-fit to compute prob of 
% 3 possible directions
function [prob] = bestFitLine (dcode)

% step 0: convert direction code to (x,y) coordinate

% to deal with dcode like [1 2 3] and [3 2 1] 
if (dcode(1)-dcode(2) == dcode(2)-dcode(3)) && abs(dcode(1)-dcode(2)) == 1
    dcode(1) = []; % just consider the last two codes
end

dLength = length(dcode);

xd(1) = 0;            % step 0.1:  initialize 1 point
yd(1) = 0;

for i=1:dLength,      % step 0.2:  convert rest of chain code
    if dcode(i) == 1,       % North
        xd(i+1) = xd(i);
        yd(i+1) = yd(i)+1;
    elseif dcode(i) == 2,   % East
        xd(i+1) = xd(i)+1;
        yd(i+1) = yd(i);
    elseif dcode(i) == 3,
        xd(i+1) = xd(i);    % South
        yd(i+1) = yd(i)-1;
    else
        xd(i+1) = xd(i)-1;  % West
        yd(i+1) = yd(i);
    end
end

% step 1: calculate least squares fitting--perpendicular offsets
% http://mathworld.wolfram.com/LeastSquaresFittingPerpendicularOffsets.html
if sum(dcode == 1) == dLength || sum(dcode == 3) == dLength,
    m = inf;

else
    xm = mean(xd);
    ym = mean(yd);
    temp1 = yd*yd' - length(xd)*ym^2 - (xd*xd' - length(xd)*xm^2);
    temp2 = length(xd)*xm*ym - xd*yd';
    if temp2 == 0;
        m = 0;
    else
        B = 0.5*temp1/temp2;
        m1 = -B + sqrt(B^2+1);  % or -B - sqrt(B^2+1)
        m2 = -B - sqrt(B^2+1);
        b1 = ym - m1*xm;
        b2 = ym - m2*xm;
        sum1 = 0;
        sum2 = 0;
        for i=1:length(xd),
            sum1 = sum1 + abs(yd(i)-(b1+m1*xd(i)))/sqrt(1+m1^2);
            sum2 = sum2 + abs(yd(i)-(b2+m2*xd(i)))/sqrt(1+m2^2);
        end
        if sum1 < sum2,
            m = m1;
        else
            m = m2;
        end
    end
end


% step 2: find angles between predicted direction & 3 poss. directions

if dcode(dLength) == 1,           % dcode: North  ******
    if m >= 0,
        ang(1) = pi-atan(m);    % left
        ang(2) = pi/2-atan(m);  % straight
        ang(3) = atan(m);       % right
    else
        ang(1) = -atan(m);      % left
        ang(2) = pi/2+atan(m);  % straight
        ang(3) = pi+atan(m);    % right
    end
    
elseif dcode(dLength) == 2,       % dcode: East ******
    ang(1) = pi/2-atan(m);  % left
    ang(2) = abs(atan(m));  % straight
    ang(3) = pi/2+atan(m);  % right
    
elseif dcode(dLength) == 3,       % dcode: South ******
    if m >= 0,
        ang(1) = pi-atan(m);    % left
        ang(2) = pi/2-atan(m);  % straight
        ang(3) = atan(m);       % right
    else
        ang(1) = -atan(m);      % left
        ang(2) = pi/2+atan(m);  % straight
        ang(3) = pi+atan(m);    % right
    end
    
else                              % dcode: West ******
    ang(1) = pi/2-atan(m);  % left
    ang(2) = abs(atan(m));  % straight
    ang(3) = pi/2+atan(m);  % right
end


% step 3: compute prob using von Mises distr.

minAng = min([ang(1) ang(2) ang(3)]);
rho = 2.0;         % set parameter rho
epsilon = 0.0000000000001;   % set parameter epsilon
%kappa = rho * cos(2*minAng) + epsilon;
kappa = rho * cos(2*minAng);
prob = zeros(1,3);
for i=1:3,
    prob(i) = exp(kappa * cos(ang(i)));
end
prob = prob/sum(prob);
