% *************************************************************
% author:   Gene Cheung                                      **
% date:     07/24/2014                                       **
% modified: 07/24/2014                                       **
% purpose:  find least-square fitting line on 2D plot        **
% *************************************************************

% given direction code, find line-of-best-fit to compute prob of 
% 3 possible directions
function [prob] = bestFitLine2 (fullDcode)

% step 0: convert direction code to (x,y) coordinate
len = length(fullDcode);
dcode = fullDcode(len-2:len);   % the last 3 data point

% to deal with dcode like [1 2 3] and [3 2 1] 
if (dcode(1)-dcode(2) == dcode(2)-dcode(3)) && abs(dcode(1)-dcode(2)) == 1
    dcode(1) = []; % just consider the last two codes
end

dLength = length(dcode);

xd(1) = 0;            % step 0.1:  initialize 1 point
yd(1) = 0;

for i=1:dLength,      % step 0.2:  convert rest of chain code
    if dcode(i) == 1,       % North
        xd(i+1) = xd(i);
        yd(i+1) = yd(i)+1;
    elseif dcode(i) == 2,   % East
        xd(i+1) = xd(i)+1;
        yd(i+1) = yd(i);
    elseif dcode(i) == 3,
        xd(i+1) = xd(i);    % South
        yd(i+1) = yd(i)-1;
    else
        xd(i+1) = xd(i)-1;  % West
        yd(i+1) = yd(i);
    end
end

% step 1: calculate least squares fitting--perpendicular offsets ********
% http://mathworld.wolfram.com/LeastSquaresFittingPerpendicularOffsets.html
if sum(dcode == 1) == dLength || sum(dcode == 3) == dLength,
    avgDiff = 0;
    m = inf;

else
    xm = mean(xd);
    ym = mean(yd);
    temp1 = yd*yd' - length(xd)*ym^2 - (xd*xd' - length(xd)*xm^2);  % eq.(18)
    temp2 = length(xd)*xm*ym - xd*yd';
    if temp2 == 0;
        avgDiff = 0;
        m = 0;
        b = ym;
    else
        B = 0.5*temp1/temp2;
        m1 = -B + sqrt(B^2+1);  % or -B - sqrt(B^2+1)
        m2 = -B - sqrt(B^2+1);
        b1 = ym - m1*xm;
        b2 = ym - m2*xm;
        sum1 = 0;
        sum2 = 0;
        for i=1:length(xd),
            sum1 = sum1 + abs(yd(i)-(b1+m1*xd(i)))/sqrt(1+m1^2);    % eq.(3)
            sum2 = sum2 + abs(yd(i)-(b2+m2*xd(i)))/sqrt(1+m2^2);
        end
        if sum1 < sum2,
            avgDiff = sum1 / length(xd);
            m = m1;
            b = b1;
        else
            avgDiff = sum2 / length(xd);
            m = m2;
            b = b2;
        end
    end
end


% step 2: find predicted direction ********
x = xd(dLength+1);   % 1 more coord than dcode
y = yd(dLength+1);
if m < inf,
    if m ~= 0,
        x2 = (y - b) / m;    % line intects w/ last coord
        y2 = m * x + b;
    else
        y2 = y;
        if dcode(dLength) == 2,
            x2 = x + 1;
        else
            x2 = x - 1;
        end
    end
else
    x2 = x;
    if dcode(dLength) == 1,
        y2 = y + 1;
    else
        y2 = y - 1;
    end
end

if dcode(dLength) == 1,                % decode: north ******
    
    if m >= 0,
        ang(1) = pi-atan(m);    % left
        ang(2) = pi/2-atan(m);  % straight
        ang(3) = atan(m);       % right
    else
        ang(1) = -atan(m);      % left
        ang(2) = pi/2+atan(m);  % straight
        ang(3) = pi+atan(m);    % right
    end
    minAng = min([ang(1) ang(2) ang(3)]);
    
    if y2 > y,
        ang(2) = 0; % ang(2)/2;
    elseif x2 < x,
        ang(1) = 0; % ang(1)/2;
    else
        ang(3) = 0; % ang(3)/2;
    end
       
    
elseif dcode(dLength) == 2,            % decode: east ******
    
    ang(1) = pi/2-atan(m);  % left
    ang(2) = abs(atan(m));  % straight
    ang(3) = pi/2+atan(m);  % right
    minAng = min([ang(1) ang(2) ang(3)]);
    
    if x2 > x,
        ang(2) = 0; % ang(2)/2;
    elseif y2 > y,
        ang(1) = 0; % ang(1)/2;
    else
        ang(3) = 0; % ang(3)/2;
    end
    
    
elseif dcode(dLength) == 3,            % decode: south ******

     if m >= 0,
        ang(1) = pi-atan(m);    % left
        ang(2) = pi/2-atan(m);  % straight
        ang(3) = atan(m);       % right
    else
        ang(1) = -atan(m);      % left
        ang(2) = pi/2+atan(m);  % straight
        ang(3) = pi+atan(m);    % right
     end
     minAng = min([ang(1) ang(2) ang(3)]);
    
    if y2 < y,
        ang(2) = 0; % ang(2)/2;
    elseif x2 > x,
        ang(1) = 0; % ang(1)/2;
    else
        ang(3) = 0; % ang(3)/2;
    end
    
    
elseif dcode(dLength) == 4,        % decode: west ******
    
    ang(1) = pi/2-atan(m);  % left
    ang(2) = abs(atan(m));  % straight
    ang(3) = pi/2+atan(m);  % right
    minAng = min([ang(1) ang(2) ang(3)]);
    
    if x2 < x,
        ang(2) = 0; % ang(2)/2;
    elseif y2 < y,
        ang(1) = 0; % ang(1)/2;
    else
        ang(3) = 0; % ang(3)/2;
    end   
    
end


% step 3: compute prob using von Mises distr. ********
% 3.1 construct history points ******
ind = 1;
xd2(ind) = 0;
yd2(ind) = 0;
hisLen = 2;
for i = len-3: -1 : max(1, len-3-hisLen),     % history length
    if fullDcode(i) == 1,          % North
        xd2(ind+1) = xd2(ind);
        yd2(ind+1) = yd2(ind)-1;
    elseif fullDcode(i) == 2,      % East
        xd2(ind+1) = xd2(ind)-1;
        yd2(ind+1) = yd2(ind);
    elseif fullDcode(i) == 3,
        xd2(ind+1) = xd2(ind);     % South
        yd2(ind+1) = yd2(ind)+1;
    else
        xd2(ind+1) = xd2(ind)+1;   % West
        yd2(ind+1) = yd2(ind);
    end
    ind = ind + 1;
end

sum1 = 0;
for i = 2 : ind,
    if m < Inf,
        sum1 = sum1 + abs(yd2(i)-(b+m*xd2(i)))/sqrt(1+m^2);  
    else
        sum1 = sum1 + abs(xd2(i) - x);
    end
end
if ind > 1,
    hisDiff = sum1 / (ind - 1);
else
    hisDiff = 0;
end

% 3.2 compute kappa and von Mises ditr. ******
rho = 2.0;
tau1 = 0.5;
tau2 = 1.0;
% kappa = rho * exp(-1*tau1*power(hisDiff - avgDiff, 2)) * exp(-1*tau2*power(avgDiff, 2));
 kappa = rho * exp(-1*tau1*abs(hisDiff - avgDiff)) * exp(-1*tau2*abs(avgDiff));
% kappa = rho * cos(2*minAng);

prob = zeros(1,3);
for i=1:3,
    prob(i) = exp(kappa * cos(ang(i)));
end
prob = prob/sum(prob);
