% ************************************************************************
% author:   Amin Zheng                                                  **
% date:     03/20/2015                                                  **
% modified: 04/23/2016  shared version                                  **
% purpose:  lossless image contour coding                               **
% ************************************************************************

tic % store the running time

% step 0: set arguments and input sequences ****

isAEC = 0; % 1: coding with AEC, 0: coding with proposed CT-LL
isBinaryCodingOfStartPnt = 0; % 1: Fixed Length Binary code, 0: M-Golomb coding

alpha = 1.0; % depth contour detector parameter, larger results in fewer contours
minLength = 10; % remove contour with length smaller than minLength
numFrameCode = 10; % number of frames to be coded

% set method to construct the context tree ***
treePrune = 'CT-LL'; % other context tree construction methods can be implemented based on this framework
alphaPrune = 0.25; % prior coefficeint: alphaPrune * ln L

% set the input type of sequences ***
typeChoices = {'mask', 'image'};
inputType = 'image';
if(sum(strcmp(inputType, typeChoices)) == 0)
    inputType = 'mask';
end

if strcmp(inputType, 'mask') == 1
    K = 1000; % maxmium number of nodes on the initial context tree
elseif strcmp(inputType, 'image') == 1
    K = 2500;
end 

% set input sequence, masks in the first line, images in the second lines ***
seqChoices = {'children_sif', 'stefan_sif', ...
    'teddy_crop_fill', 'Laundry_disp1_crop_fill'};
seqName = 'teddy_crop_fill';

if(sum(strcmp(seqName, seqChoices)) == 0)
    seqName = 'children_sif';
end

% set path of input sequences ***
if strcmp(inputType, 'mask') == 1
    fRoot0 = ['input\mask\', seqName, '\f00'];
    fRoot = ['input\mask\', seqName, '\f0'];
    fEx = '.png';
    
elseif strcmp(inputType, 'image') == 1
    f1 = ['input\image\', seqName, '.png'];
    fStat = 'output\edgeCost.txt';
    
    numFrameCode = 1;
end
contextNum = zeros(numFrameCode,1); % store the context number of each frame

% set the online training images ***
trainImgs1 = {'Wood2_disp1_crop_fill', 'Aloe_disp5_crop_fill','Lampshade1_disp1_crop_fill','Art_disp1_crop_fill'};
trainImgs2 = {'Lampshade2_disp5_crop_fill', 'Baby1_disp1_crop_fill','Dolls_disp1_crop_fill','Books_disp1_crop_fill'};
trainImgs12 ={'Wood2_disp1_crop_fill', 'Aloe_disp5_crop_fill','Lampshade1_disp1_crop_fill','Art_disp1_crop_fill',...
    'Lampshade2_disp5_crop_fill', 'Baby1_disp1_crop_fill','Dolls_disp1_crop_fill','Books_disp1_crop_fill'};
fRootTrain = 'input\train\Middleburry\';

trainSet = 2; % 1: train set 1, 2: train set 2, 0: train set 12
if trainSet == 0
    trainImgs = trainImgs12;
elseif trainSet == 1
    trainImgs = trainImgs1;
elseif trainSet == 2
    trainImgs = trainImgs2;
end

% step 1: if input is image, get the conditional probability and optimal context tree for coding ***
if strcmp(inputType, 'image') == 1
    % step 1.1: get the conditional probability ***
    imgContSet = cell(length(trainImgs), 1);
    
    for i = 1:length(trainImgs)
        fTrainName = [fRootTrain, trainImgs{i}, '.png'];
        imTrain = imread(fTrainName);
        imTrain = double(imTrain(:,:,1));   % take the first component if png format
        
        % contour dectection and convert detected contours into DCC strings
        [~, dcodeTrain, ~] = getContourStartPnt(imTrain, alpha, minLength);
        
        imgContSet{i} = dcodeTrain;
    end
    
    % get conditional probabilities for nodes on the initial context tree ***
    [countStateSpace, orderRange] = probTrainMCT(0, length(trainImgs), imgContSet, treePrune, K);
    
    % for debug, record the coding information of different contexts ***
    contextBits = zeros(orderRange, 3^orderRange, 2);
        
    % step 1.2: get optimal context tree from the initial context tree ***
    [treeHeight, maxOrder, orgTreeHeight, ctxNum] = pruningMCT(countStateSpace, treePrune, alphaPrune);
    
    contextNum(1) = ctxNum;
end

perf = zeros(numFrameCode,1); % store the bits per symbol of each frame
maxOrderVec = zeros(numFrameCode, 1); % store maximum depth of context tree of each frame
bitsStartPnt = zeros(numFrameCode, 1); % store coding bits of starting points of each frame
symCount = zeros(numFrameCode, 1);   % store number of input symbols of each frame
bCount = zeros(numFrameCode, 1);     % store total coding bits of each frame
contourNum = zeros(numFrameCode, 1); % store contour number of each frame


for frame = 1:numFrameCode,
    
    if ~strcmp(inputType, 'image') == 1
        if frame < 10,
            fName = strcat(fRoot0, int2str(frame));
        else
            fName = strcat(fRoot, int2str(frame));
        end
        f1 = strcat(fName, fEx);
    end
    
    im1 = imread(f1);
    im1 = double(im1(:,:,1));   % take the first component if png format
    [imRow, imCol] = size(im1);
    
    % step 2: contour detection ***
    % step 2.1: get the contours and starting points ****
    [startPt, dcode] = getContourStartPnt(im1, alpha, minLength);
    
    [dRow, dCol] = size(dcode);
    flag = ones(dRow, 1); % for multiple tree part
    
    % step 2.2: compressing the starting points ***
    % remark: don't need to code startPt.hGrad, because it can be inferred from the first contour pixel ***
    if isBinaryCodingOfStartPnt == 1 
        bitsStartPnt(frame) = dRow * (ceil(log2(imRow)) + ceil(log2(imCol)));
        startPntMethod = 'Binary';
    else
        bitsStartPnt(frame) = golombCodeStartPnt(imRow, imCol, startPt, dRow);
        startPntMethod = 'M-Golomb';
    end
    
    % step 3: compute actual total coding bits *******
    bCountPerCont = zeros(1, dRow);
    bitPerSymbPerCont = zeros(2, dRow);
    
    for i=1:dRow, % for each contour
        
        [val, ind] = min(dcode(i,:));
        dcode1 = dcode(i,1:ind-1);         % read the i-th contour
        if (frame == 1 && numFrameCode > 1) || isAEC == 1 % first frame use AEC
            [buffer, bCount1] = aec(dcode1);
        else
            % coding using optimal context tree ***
            [buffer, bCount1, contextBits] = aecMCT (maxOrder, dcode1, countStateSpace, treeHeight, contextBits);
        end
        [base2,symNum] = deci2bin(ind-1);  % represent the length of the contour in binary
        % fprintf('Bits per symbol of contour %d: %d\n', i, bCount1/(ind-1));
        
        bCount1 = bCount1 + symNum;        % add the bits to record the length of the contour
        
        bCountPerCont(i) = bCount1;
        bitPerSymbPerCont(1, i) = bCount1 / (ind - 1);
        bitPerSymbPerCont(2, i) = ind - 1;
        
        % step 3.2: accumulate bits and symdbols ******
        bCount(frame) = bCount(frame) + bCount1;
        symCount(frame) = symCount(frame) + ind-1;
    end
    bCount(frame) = bCount(frame) + bitsStartPnt(frame);
    perf(frame) = bCount(frame) / symCount(frame); % compute number of bits per symbol
    
    contourNum(frame) = dRow;
    
    % step 4: if input is mask, get conditional probabilities and context tree for coding ***
    if strcmp(inputType, 'image') ~= 1
        % step 4.1: get conditional probabilities of nodes on the initial context tree ***
        if frame > 1 % two previous frames as training images
            [countStateSpace, orderRange] = probTrainMCT(0, 2, {dcode, prevDcode}, treePrune, K);
        else
            [countStateSpace, orderRange] = probTrainMCT(0, 1, {dcode}, treePrune, K);
        end
        
        % step 4.2: get optimal context tree ***
        [treeHeight, maxOrder, orgTreeHeight, ctxNum] = pruningMCT(countStateSpace, treePrune, alphaPrune);
        
        contextNum(frame+1) = ctxNum;
        
        % for debug, record the coding information of different contexts ***
        contextBits = zeros(orderRange, 3^orderRange, 2);
        
        % preserve previous information ***
        prevStartPt = startPt;  
        prevDcode = dcode;
        prevFlag = flag;
    end
    maxOrderVec(frame) = maxOrder;
    
    % step 5: output computed contour as gradient maps ********
    [hGrad2, vGrad2] = contour2edgeMap(startPt, dcode, imRow, imCol);
    
    % printEdgeMap(hGrad2, vGrad2, fName, horizon); ******
    %     printContour(startPt, dcode, bCountPerCont, fName, horizon);
    
    % step 6: visual check ********
    figure(2);
    displayEdgeMap(hGrad2, vGrad2, im1);
    
end % for frame = 1:numFrameCode,

% step 7: store the coding bits ***

% step 7.1: set path of output ***
if strcmp(inputType, 'mask') == 1
    fRoot = ['output\mask\', seqName, '\'];
elseif strcmp(inputType, 'image') == 1
    fRoot = ['output\image\', seqName, '_'];
end

if isBinaryCodingOfStartPnt == 0
    methodOfStartPnt = 'MGolomb';
else
    methodOfStartPnt = 'Binary';
end

if isAEC == 1
    fOutName = [fRoot, 'AEC', '_', methodOfStartPnt, '.txt'];
elseif isAEC == 0
    if strcmp(inputType, 'image') == 1
        fOutName = [fRoot, treePrune, '_TrainSet', num2str(trainSet),'_',methodOfStartPnt, '.txt'];
    else
        fOutName = [fRoot, treePrune, '_', methodOfStartPnt, '.txt'];
    end
end

bitsOutFid = fopen(fOutName, 'wt');

% step 7.2: store the overall information ***
totalBitsStartPnt = sum(bitsStartPnt ); % total starting points bits
totalBCount = sum(bCount); % total bits
totalSymCount = sum(symCount); % total number of symbols
aveMaxOrderVec = mean(maxOrderVec);

if numFrameCode == 1
    aveBitsPerSym = totalBCount / totalSymCount;
else
    % do not consider the first frame because it is coded with AEC for all methods *** 
    aveBitsPerSym = sum(bCount(2:end)) / sum(symCount(2:end));
end

totalContourNum = sum(contourNum);

% step 7.3: store information of each frame ***
if isAEC == 1 % AEC, no max order
    fprintf(bitsOutFid, 'Frame Number: %d, Start Points Coding: %s\n\n',...
        numFrameCode, startPntMethod);
    fprintf(bitsOutFid, 'Total Bits: %d, Total Contours: %d, StartPnt Bits: %d, Total Symbols: %d, Average Bits per Symbol: %.3f\n\n',...
        totalBCount, totalContourNum, totalBitsStartPnt, totalSymCount, aveBitsPerSym);
    fprintf(bitsOutFid, 'Frame:    TotalBits    ContNum    StartBits    TotalSym    BitsPerSym\n');
    
    for frame = 1:numFrameCode
        fprintf(bitsOutFid, '%-10d%-13d%-11d%-13d%-12d%-14.3f\n',frame, bCount(frame),...
            contourNum(frame), bitsStartPnt(frame), symCount(frame), perf(frame));
    end
elseif isAEC == 0 % CT-LL
    fprintf(bitsOutFid, 'Frame Number: %d, Start Points Coding: %s, Pruning Method: %s, Alpha: %.3f, K: %d\n\n',...
        numFrameCode, startPntMethod, treePrune, alphaPrune, K);
    fprintf(bitsOutFid, 'Total Bits: %d, Total Contours: %d, StartPnt Bits: %d, Total Symbols: %d, Average Bits per Symbol: %.3f, Average Maximum Order: %.3f\n\n',...
        totalBCount, totalContourNum, totalBitsStartPnt, totalSymCount, aveBitsPerSym, aveMaxOrderVec);
    fprintf(bitsOutFid, 'Frame:    TotalBits    ContNum    StartBits    TotalSym    BitsPerSym    MaxOrder\n');
    
    for frame = 1:numFrameCode
        fprintf(bitsOutFid, '%-10d%-13d%-11d%-13d%-12d%-14.3f%-8d\n',frame, bCount(frame),...
            contourNum(frame), bitsStartPnt(frame), symCount(frame), perf(frame), maxOrderVec(frame));
    end
end

fclose(bitsOutFid);

toc
