% ************************************************************************
% author:   Amin Zheng                                                  **
% date:     06/23/2015                                                  **
% modified:                                                             **
% purpose:  get the pruned tree with MAP method                         **
% ************************************************************************

function [treeHeight, prunedSize, orgTreeHeight] = pruningMAP(countStateSpace, maxOrder, alphaPrune)

% step 1: paramters and variables initialization
% step 1.1: parameter initialization

count1order = countStateSpace(1,:,:);% get the total number of samples ******
totalNumData = sum(count1order(:));

% parameter for shape prior
alpha = alphaPrune*log(totalNumData);

% prob of contex with occurrences <= thNumData is calculated based on its parent ***
thNumData = 0;

% offset to avoid zero probability
deltaOffset = 1/3; 

% step 1.2 variables initilization
treeHeight = zeros(maxOrder, 3^maxOrder);

% intermediate variables
stat2IndVec = zeros(maxOrder, maxOrder);
for i = 1:maxOrder
    stat2IndVec(i, 1:i) = 3.^(0:i-1);
end

isSplit = -1 * ones(maxOrder - 1, 3^(maxOrder - 1));
u = zeros(maxOrder, 1); % store the conditional states
initDepth = 0; % the root

[mapCost] = recurMAP(initDepth, u, count1order(:));
    function [mapCost] = recurMAP(depth, u, count)
        % get the map cost and number of data sample given the conditinal states ***
        if depth == 0
            nonSplitCost = realmax('single');
        else
            [nonSplitLikelihood, nonSplitPrior] = getMAP(depth, u, count);
            nonSplitCost = nonSplitLikelihood + nonSplitPrior;
            stateIndex = stat2IndVec(depth, 1:depth)* (u(depth:-1:1) - 1) + 1;
        end
               
        % stop at the maximum depth ******
        if depth == maxOrder
            mapCost = nonSplitCost;
        else
            u(depth + 1) = 1;
            stateIndexL = stat2IndVec(depth+1, 1:depth+1)* (u(depth+1:-1:1) - 1) + 1;
            stateIndexS = stateIndexL + 1;
            stateIndexR = stateIndexL + 2;
            
            countL = countStateSpace(depth+1, stateIndexL, :);countL = countL(:);
            countS = countStateSpace(depth+1, stateIndexS, :);countS = countS(:);
            countR = countStateSpace(depth+1, stateIndexR, :);countR = countR(:);
            
            thNumL = 0;
            thNumS = 0;
            thNumR = 0;
            
            if sum(countL) > thNumData
                u(depth + 1) = 1;
                [splitCostL] = recurMAP(depth + 1, u, countL);
                thNumL = sum(countL);
            else
                splitCostL = 0;
            end
            
            if sum(countS) > thNumData
                u(depth + 1) = 2;
                [splitCostS] = recurMAP(depth + 1, u, countS);
                thNumS = sum(countS);
            else
                splitCostS = 0;
            end
            
            if sum(countR) > thNumData
                u(depth + 1) = 3;
                [splitCostR] = recurMAP(depth + 1, u, countR);
                thNumR = sum(countR);
            else
                splitCostR = 0;
            end
                        
            if depth == 0 % always divide for root
                mapCost = splitCostL + splitCostS + splitCostR;
            else
                % Since some contexts are cut off, the second term is to make sure...
                % that the total numbers of symbols of parent and children are the same ***
                splitCost = (splitCostL + splitCostS + splitCostR) + ...
                    (sum(count) - thNumL - thNumR - thNumS)/sum(count)*nonSplitLikelihood;
               
                % stop divding when children have no occurrences or cost increase
                if (sum(countL) <= thNumData && sum(countS) <= thNumData &&sum(countR) <= thNumData )...
                        || splitCost >= nonSplitCost
                    isSplit(depth, stateIndex) = 0;
                    mapCost = nonSplitCost;
                else
                    isSplit(depth, stateIndex) = 1;
                    mapCost = splitCost;
                end
            end
            
        end
    end

    function [nonSplitLikelihood, nonSplitPrior] = getMAP(depth, u, count)      
        numData = sum(count);     
        prob = (count + deltaOffset) / (3*deltaOffset + numData); % add offset to avoid zero
        
        nonSplitLikelihood = -numData * (prob(1)*log(prob(1))+prob(2)*log(prob(2))...
            + prob(3)*log(prob(3)));
        nonSplitPrior = alpha * (max(0, directionTurn(depth, u)));      
    end

    function [num] = directionTurn(depth, u)
        
        if depth >= 2
            xd = zeros(depth+1,1);
            yd = zeros(depth+1,1);
            
            orient = 0; % 0,1,2,3 ~ north, east, south, west            
            yd(2) = yd(1) + 1;
            
            for indCor = 2:depth,
                if u(indCor) == 1     % left
                    orient = mod(orient-1, 4);
                elseif u(indCor) == 3 % rigth
                    orient = mod(orient+1, 4);
                end
                
                if orient == 0,       % North
                    xd(indCor+1) = xd(indCor);
                    yd(indCor+1) = yd(indCor)+1;
                elseif orient == 1,   % East
                    xd(indCor+1) = xd(indCor)+1;
                    yd(indCor+1) = yd(indCor);
                elseif orient == 2,
                    xd(indCor+1) = xd(indCor);  % South
                    yd(indCor+1) = yd(indCor)-1;
                else                  % West
                    xd(indCor+1) = xd(indCor)-1;
                    yd(indCor+1) = yd(indCor);
                end
            end
            
            dist = zeros(depth-1, 1);
            for indCor = 2:depth,
                dist(indCor-1) = abs((xd(indCor)-xd(1))*(yd(depth+1)-yd(1))-...
                    (yd(indCor)-yd(1))*(xd(depth+1)-xd(1)))/...
                    sqrt((xd(depth+1)-xd(1))^2+(yd(depth+1)-yd(1))^2);
            end
            num = max(dist);
        else
            num = 0;
        end
       
    end

% get the tree height from the split tree: isSplit ******
tempSplitIndex = zeros(maxOrder - 1, 1);

for i = 1:3^(maxOrder - 1)
    
    curI = i;
    curDepth = maxOrder - 1;
    tempSplitIndex(curDepth) = isSplit(curDepth, curI);
    
    while  curDepth > 1
        curDepth = curDepth - 1;
        curI = floor((curI-1)/3.0)+1;
        tempSplitIndex(curDepth) = isSplit(curDepth, curI);
    end
    
    % search split tree from top to down ******
    ind = find(tempSplitIndex < 1);
    if isempty(ind)
        height = maxOrder;
    else
        height = ind(1);
    end
    
    % detect the number of data sample of current height ******
    childL = (i - 1)*3 + 1; % left child of node i 
    childS = (i - 1)*3 + 2;
    childR = (i - 1)*3 + 3;
    
    treeHeight(maxOrder, childL) = height;
    treeHeight(maxOrder, childS) = height;
    treeHeight(maxOrder, childR) = height;
end

% get the treeHeight for trees with depth < maxOrder ***
for i = 1:3^maxOrder
    height = treeHeight(maxOrder, i);    
    for j = maxOrder - 1:-1:1
        if mod(i, 3^(maxOrder - j)) == 0
            tempI = i / (3^(maxOrder - j));
            
            treeHeight(j, tempI) = min(height, j);
        end
    end
end
    
orgTreeHeight = treeHeight;

% prob of contex with occurrences <= thNumData is calculated based on its parent ***
prunedSize = zeros(maxOrder, 1);

for iOrder = 1:maxOrder
    
    terminTree = -1*ones(iOrder, 3^iOrder);
    for i = 1:3^iOrder
        height = treeHeight(iOrder, i);
        curI = i;
        curDepth = iOrder;
        
        while curDepth > height
            curDepth = curDepth - 1;
            curI = floor((curI-1)/3.0)+1;
        end
        
        tempCount = countStateSpace(curDepth, curI, :);
        tempCount = tempCount(:);
        
        if sum(tempCount) <= thNumData
            treeHeight(iOrder, i) = height - 1;
%             terminTree(height - 1, floor((curI-1)/3.0)+1) = 1; % parent is also a context
            terminTree(height, curI) = 1; % contex with occurrences <= thNumData is also a context
        else
            terminTree(height, curI) = 1;
        end
    end
    prunedSize(iOrder) = numel(find(terminTree == 1));
end

end